#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['fastapi_users_db_sqlmodel']

package_data = \
{'': ['*']}

install_requires = \
['fastapi-users >= 7.0.0', 'sqlmodel >=0.0.4,<0.1.0']

setup(name='fastapi-users-db-sqlmodel',
      version='0.0.4',
      description='FastAPI Users database adapter for SQLModel.',
      author='François Voron',
      author_email='fvoron@gmail.com',
      url='https://github.com/fastapi-users/fastapi-users-db-sqlmodel',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.7',
     )
